/*
 * Copyright (c) 2012-2014 Wind River Systems, Inc.
 *
 * SPDX-License-Identifier: Apache-2.0
 */
 
#if CONFIG_BOARD_CUSTOM_NRF5340_CPUAPP

#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/spi.h>
#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include <zephyr/sys/printk.h>
#include <zephyr/kernel.h>
#include "adc.h"
#include "error.h"

// clock frequency
#define AD7124_FCLK_FULL_POWER 614400
#define AD7124_FCLK_MID_POWER  153600
#define AD7124_FCLK_LOW_POWER  76800

// register address
#define AD7124_REG_ADDR_COMMS        0x00
#define AD7124_REG_ADDR_STATUS       0x00
#define AD7124_REG_ADDR_ADC_CONTROL  0x01
#define AD7124_REG_ADDR_DATA         0x02
#define AD7124_REG_ADDR_IO_CONTROL_1 0x03
#define AD7124_REG_ADDR_IO_CONTROL_2 0x04
#define AD7124_REG_ADDR_ID           0x05
#define AD7124_REG_ADDR_ERROR        0x06
#define AD7124_REG_ADDR_ERROR_EN     0x07
#define AD7124_REG_ADDR_MCLK_COUNT   0x08
#define AD7124_REG_ADDR_CHANNEL(x)   (0x09 + (x))
#define AD7124_REG_ADDR_CONFIG(x)    (0x19 + (x))
#define AD7124_REG_ADDR_FILTER(x)    (0x21 + (x))
#define AD7124_REG_ADDR_OFFSET(x)    (0x29 + (x))
#define AD7124_REG_ADDR_GAIN(x)      (0x31 + (x))

// register bit
#define AD7124_REG_COMMS_BIT_RS                          0
#define AD7124_REG_COMMS_BIT_RW                          6
#define AD7124_REG_COMMS_BIT_WEN                         7
#define AD7124_REG_STATUS_BIT_CH_ACTIVE                  0
#define AD7124_REG_STATUS_BIT_POR_FLAG                   4
#define AD7124_REG_STATUS_BIT_ERROR_FLAG                 6
#define AD7124_REG_STATUS_BIT_RDY                        7
#define AD7124_REG_ADC_CONTROL_BIT_CLK_SEL               0
#define AD7124_REG_ADC_CONTROL_BIT_MODE                  2
#define AD7124_REG_ADC_CONTROL_BIT_POWER_MODE            6
#define AD7124_REG_ADC_CONTROL_BIT_REF_EN                8
#define AD7124_REG_ADC_CONTROL_BIT_CS_EN                 9
#define AD7124_REG_ADC_CONTROL_BIT_DATA_STATUS           10
#define AD7124_REG_ADC_CONTROL_BIT_CONT_READ             11
#define AD7124_REG_ADC_CONTROL_BIT_DOUT_RDY_DEL          12
#define AD7124_REG_IO_CONTROL_1_BIT_IOUT0_CH             0
#define AD7124_REG_IO_CONTROL_1_BIT_IOUT1_CH             4
#define AD7124_REG_IO_CONTROL_1_BIT_IOUT0                8
#define AD7124_REG_IO_CONTROL_1_BIT_IOUT1                11
#define AD7124_REG_IO_CONTROL_1_BIT_PDSW                 15
#define AD7124_REG_IO_CONTROL_1_BIT_GPIO_CTRL1           18
#define AD7124_REG_IO_CONTROL_1_BIT_GPIO_CTRL2           19
#define AD7124_REG_IO_CONTROL_1_BIT_GPIO_DAT1            22
#define AD7124_REG_IO_CONTROL_1_BIT_GPIO_DAT2            23
#define AD7124_REG_IO_CONTROL_2_BIT_VBIAS0               0
#define AD7124_REG_IO_CONTROL_2_BIT_VBIAS1               1
#define AD7124_REG_IO_CONTROL_2_BIT_VBIAS2               4
#define AD7124_REG_IO_CONTROL_2_BIT_VBIAS3               5
#define AD7124_REG_IO_CONTROL_2_BIT_VBIAS4               10
#define AD7124_REG_IO_CONTROL_2_BIT_VBIAS5               11
#define AD7124_REG_IO_CONTROL_2_BIT_VBIAS6               14
#define AD7124_REG_IO_CONTROL_2_BIT_VBIAS7               15
#define AD7124_REG_ID_BIT_SILICON_REVISION               0
#define AD7124_REG_ID_BIT_DEVICE_ID                      4
#define AD7124_REG_ERROR_BIT_ROM_CRC_ERR                 0
#define AD7124_REG_ERROR_BIT_MM_CRC_ERR                  1
#define AD7124_REG_ERROR_BIT_SPI_CRC_ERR                 2
#define AD7124_REG_ERROR_BIT_SPI_WRITE_ERR               3
#define AD7124_REG_ERROR_BIT_SPI_READ_ERR                4
#define AD7124_REG_ERROR_BIT_SPI_SCLK_CNT_ERR            5
#define AD7124_REG_ERROR_BIT_SPI_IGNORE_ERR              6
#define AD7124_REG_ERROR_BIT_ALDO_PSM_ERR                7
#define AD7124_REG_ERROR_BIT_DLDO_PSM_ERR                9
#define AD7124_REG_ERROR_BIT_REF_DET_ERR                 11
#define AD7124_REG_ERROR_BIT_AINM_UV_ERR                 12
#define AD7124_REG_ERROR_BIT_AINM_OV_ERR                 13
#define AD7124_REG_ERROR_BIT_AINP_UV_ERR                 14
#define AD7124_REG_ERROR_BIT_AINP_OV_ERR                 15
#define AD7124_REG_ERROR_BIT_ADC_SAT_ERR                 16
#define AD7124_REG_ERROR_BIT_ADC_CONV_ERR                17
#define AD7124_REG_ERROR_BIT_ADC_CAL_ERR                 18
#define AD7124_REG_ERROR_BIT_LDO_CAP_ERR                 19
#define AD7124_REG_ERROR_EN_BIT_ROM_CRC_ERR_EN           0
#define AD7124_REG_ERROR_EN_BIT_MM_CRC_ERR_EN            1
#define AD7124_REG_ERROR_EN_BIT_SPI_CRC_ERR_EN           2
#define AD7124_REG_ERROR_EN_BIT_SPI_WRITE_ERR_EN         3
#define AD7124_REG_ERROR_EN_BIT_SPI_READ_ERR_EN          4
#define AD7124_REG_ERROR_EN_BIT_SPI_SCLK_CNT_ERR_EN      5
#define AD7124_REG_ERROR_EN_BIT_SPI_IGNORE_ERR_EN        6
#define AD7124_REG_ERROR_EN_BIT_ALDO_PSM_ERR_EN          7
#define AD7124_REG_ERROR_EN_BIT_ALDO_PSM_TRIP_TEST_EN    8
#define AD7124_REG_ERROR_EN_BIT_DLDO_PSM_ERR_EN          9
#define AD7124_REG_ERROR_EN_BIT_DLDO_PSM_TRIP_TEST_EN    10
#define AD7124_REG_ERROR_EN_BIT_REF_DET_ERR_EN           11
#define AD7124_REG_ERROR_EN_BIT_AINM_UV_ERR_EN           12
#define AD7124_REG_ERROR_EN_BIT_AINM_OV_ERR_EN           13
#define AD7124_REG_ERROR_EN_BIT_AINP_UV_ERR_EN           14
#define AD7124_REG_ERROR_EN_BIT_AINP_OV_ERR_EN           15
#define AD7124_REG_ERROR_EN_BIT_ADC_SAT_ERR_EN           16
#define AD7124_REG_ERROR_EN_BIT_ADC_CONV_ERR_EN          17
#define AD7124_REG_ERROR_EN_BIT_ADC_CAL_ERR_EN           18
#define AD7124_REG_ERROR_EN_BIT_LDO_CAP_CHK              19
#define AD7124_REG_ERROR_EN_BIT_LDO_CAP_CHK_TEST_EN      21
#define AD7124_REG_ERROR_EN_BIT_MCLK_CNT_EN              22
#define AD7124_REG_CHANNEL_BIT_AINM                      0
#define AD7124_REG_CHANNEL_BIT_AINP                      5
#define AD7124_REG_CHANNEL_BIT_SETUP                     12
#define AD7124_REG_CHANNEL_BIT_ENABLE                    15
#define AD7124_REG_CONFIG_BIT_PGA                        0
#define AD7124_REG_CONFIG_BIT_REF_SEL                    3
#define AD7124_REG_CONFIG_BIT_AIN_BUFM                   5
#define AD7124_REG_CONFIG_BIT_AIN_BUFP                   6
#define AD7124_REG_CONFIG_BIT_REF_BUFM                   7
#define AD7124_REG_CONFIG_BIT_REF_BUFP                   8
#define AD7124_REG_CONFIG_BIT_BURNOUT                    9
#define AD7124_REG_CONFIG_BIT_BIPOLAR                    11
#define AD7124_REG_FILTER_BIT_FS                         0
#define AD7124_REG_FILTER_BIT_SINGLE_CYCLE               16
#define AD7124_REG_FILTER_BIT_POST_FILTER                17
#define AD7124_REG_FILTER_BIT_REJ60                      20
#define AD7124_REG_FILTER_BIT_FILTER                     21

// register length
#define AD7124_REG_LEN_COMMS    1
#define AD7124_REG_LEN_STATUS   1
#define AD7124_REG_LEN_ADC_CTRL 2
#define AD7124_REG_LEN_ID       1
#define AD7124_REG_LEN_CONFIG   2
#define AD7124_REG_LEN_FILTER   3
#define AD7124_REG_LEN_OFFSET   3
#define AD7124_REG_LEN_GAIN     3

// output date rate
#define AD7124_OUTPUT_DATA_RATE_5    5
#define AD7124_OUTPUT_DATA_RATE_10   10
#define AD7124_OUTPUT_DATA_RATE_20   20
#define AD7124_OUTPUT_DATA_RATE_40   40
#define AD7124_OUTPUT_DATA_RATE_50   50
#define AD7124_OUTPUT_DATA_RATE_60   60
#define AD7124_OUTPUT_DATA_RATE_80   80
#define AD7124_OUTPUT_DATA_RATE_160  160
#define AD7124_OUTPUT_DATA_RATE_300  300
#define AD7124_OUTPUT_DATA_RATE_600  600
#define AD7124_OUTPUT_DATA_RATE_1200 1200
#define AD7124_OUTPUT_DATA_RATE_2400 2400

// communication register (addr=0x00, len=8, W)
#define AD7124_COMMS_WEN_MSK BIT(7)
#define AD7124_COMMS_WEN(x)  FIELD_PREP(AD7124_COMMS_WEN_MSK, x)
#define AD7124_COMMS_RW_MSK  BIT(6)
#define AD7124_COMMS_RW(x)   FIELD_PREP(AD7124_COMMS_RW_MSK, x)
#define AD7124_COMMS_RS_MSK  GENMASK(5, 0)
#define AD7124_COMMS_RS(x)   FIELD_PREP(AD7124_COMMS_RS_MSK, x)

// status register (addr=0x00, len=8, R)
#define AD7124_STATUS_RDY_MSK        BIT(7)
#define AD7124_STATUS_RDY(x)         FIELD_PREP(AD7124_STATUS_RDY_MSK, x)
#define AD7124_STATUS_ERROR_FLAG_MSK BIT(6)
#define AD7124_STATUS_ERROR_FLAG(x)  FIELD_PREP(AD7124_STATUS_ERROR_FLAG_MSK, x)
#define AD7124_STATUS_POR_FLAG_MSK   BIT(4)
#define AD7124_STATUS_POR_FLAG(x)    FIELD_PREP(AD7124_STATUS_POR_FLAG_MSK, x)
#define AD7124_STATUS_CH_ACTIVE_MSK  GENMASK(3, 0)
#define AD7124_STATUS_CH_ACTIVE(x)   FIELD_PREP(AD7124_STATUS_CH_ACTIVE_MSK, x)

// ADC_CONTROL register (addr=0x01, len=16)
#define AD7124_ADC_CTRL_DOUT_RDY_DEL_MSK BIT(12)
#define AD7124_ADC_CTRL_DOUT_RDY_DEL(x)  FIELD_PREP(AD7124_ADC_CTRL_DOUT_RDY_DEL_MSK, x)
#define AD7124_ADC_CTRL_CONT_READ_MSK    BIT(11)
#define AD7124_ADC_CTRL_CONT_READ(x)     FIELD_PREP(AD7124_ADC_CTRL_CONT_READ_MSK, x)
#define AD7124_ADC_CTRL_DATA_STATUS_MSK  BIT(10)
#define AD7124_ADC_CTRL_DATA_STATUS(x)   FIELD_PREP(AD7124_ADC_CTRL_DATA_STATUS_MSK, x)
#define AD7124_ADC_CTRL_CS_EN_MSK        BIT(9)
#define AD7124_ADC_CTRL_CS_EN(x)         FIELD_PREP(AD7124_ADC_CTRL_CS_EN_MSK, x)
#define AD7124_ADC_CTRL_REF_EN_MSK       BIT(8)
#define AD7124_ADC_CTRL_REF_EN(x)        FIELD_PREP(AD7124_ADC_CTRL_REF_EN_MSK, x)
#define AD7124_ADC_CTRL_POWER_MODE_MSK   GENMASK(7, 6)
#define AD7124_ADC_CTRL_POWER_MODE(x)    FIELD_PREP(AD7124_ADC_CTRL_POWER_MODE_MSK, x)
#define AD7124_ADC_CTRL_MODE_MSK         GENMASK(5, 2)
#define AD7124_ADC_CTRL_MODE(x)          FIELD_PREP(AD7124_ADC_CTRL_MODE_MSK, x)
#define AD7124_ADC_CTRL_CLK_SEL_MSK      GENMASK(1, 0)
#define AD7124_ADC_CTRL_CLK_SEL(x)       FIELD_PREP(AD7124_ADC_CTRL_CLK_SEL_MSK, x)

// IO_CONTROL_1 register (addr=0x03, len=24)
#define AD7124_IO_CONTROL_1_GPIO_DAT2_MSK  BIT(23)
#define AD7124_IO_CONTROL_1_GPIO_DAT2      FIELD_PREP(AD7124_IO_CONTROL_1_GPIO_DAT2_MSK, x)
#define AD7124_IO_CONTROL_1_GPIO_DAT1_MSK  BIT(22)
#define AD7124_IO_CONTROL_1_GPIO_DAT1      FIELD_PREP(AD7124_IO_CONTROL_1_GPIO_DAT1_MSK, x)
#define AD7124_IO_CONTROL_1_GPIO_CTRL2_MSK BIT(19)
#define AD7124_IO_CONTROL_1_GPIO_CTRL2     FIELD_PREP(AD7124_IO_CONTROL_1_GPIO_CTRL2_MSK, x)
#define AD7124_IO_CONTROL_1_GPIO_CTRL1_MSK BIT(18)
#define AD7124_IO_CONTROL_1_GPIO_CTRL1     FIELD_PREP(AD7124_IO_CONTROL_1_GPIO_CTRL1_MSK, x)
#define AD7124_IO_CONTROL_1_PDSW_MSK       BIT(15)
#define AD7124_IO_CONTROL_1_PDSW           FIELD_PREP(AD7124_IO_CONTROL_1_PDSW_MSK, x)
#define AD7124_IO_CONTROL_1_IOUT1_MSK      GENMASK(13, 11)
#define AD7124_IO_CONTROL_1_IOUT1(x)       FIELD_PREP(AD7124_IO_CONTROL_1_IOUT1_MSK, x)
#define AD7124_IO_CONTROL_1_IOUT0_MSK      GENMASK(10, 8)
#define AD7124_IO_CONTROL_1_IOUT0(x)       FIELD_PREP(AD7124_IO_CONTROL_1_IOUT0_MSK, x)
#define AD7124_IO_CONTROL_1_IOUT1_CH_MSK   GENMASK(7, 4)
#define AD7124_IO_CONTROL_1_IOUT1_CH(x)    FIELD_PREP(AD7124_IO_CONTROL_1_IOUT1_CH_MSK, x)
#define AD7124_IO_CONTROL_1_IOUT0_CH_MSK   GENMASK(3, 0)
#define AD7124_IO_CONTROL_1_IOUT0_CH(x)    FIELD_PREP(AD7124_IO_CONTROL_1_IOUT0_CH_MSK, x)

// IO_CONTROL_2 register (addr=0x04, len=16)
#define AD7124_IO_CONTROL_2_VBIAS7_MSK  BIT(15)
#define AD7124_IO_CONTROL_2_VBIAS7      FIELD_PREP(AD7124_IO_CONTROL_2_VBIAS7_MSK, x)
#define AD7124_IO_CONTROL_2_VBIAS6_MSK  BIT(14)
#define AD7124_IO_CONTROL_2_VBIAS6      FIELD_PREP(AD7124_IO_CONTROL_2_VBIAS6_MSK, x)
#define AD7124_IO_CONTROL_2_VBIAS5_MSK  BIT(11)
#define AD7124_IO_CONTROL_2_VBIAS5      FIELD_PREP(AD7124_IO_CONTROL_2_VBIAS5_MSK, x)
#define AD7124_IO_CONTROL_2_VBIAS4_MSK  BIT(10)
#define AD7124_IO_CONTROL_2_VBIAS4      FIELD_PREP(AD7124_IO_CONTROL_2_VBIAS4_MSK, x)
#define AD7124_IO_CONTROL_2_VBIAS3_MSK  BIT(5)
#define AD7124_IO_CONTROL_2_VBIAS3      FIELD_PREP(AD7124_IO_CONTROL_2_VBIAS3_MSK, x)
#define AD7124_IO_CONTROL_2_VBIAS2_MSK  BIT(4)
#define AD7124_IO_CONTROL_2_VBIAS2      FIELD_PREP(AD7124_IO_CONTROL_2_VBIAS2_MSK, x)
#define AD7124_IO_CONTROL_2_VBIAS1_MSK  BIT(1)
#define AD7124_IO_CONTROL_2_VBIAS1      FIELD_PREP(AD7124_IO_CONTROL_2_VBIAS1_MSK, x)
#define AD7124_IO_CONTROL_2_VBIAS0_MSK  BIT(0)
#define AD7124_IO_CONTROL_2_VBIAS0      FIELD_PREP(AD7124_IO_CONTROL_2_VBIAS0_MSK, x)

// ID register (addr=0x05, len=8)
#define AD7124_4_CHIPID              0x0
#define AD7124_8_CHIPID              0x1
#define AD7124_ID_DEVICE_ID_MSK      GENMASK(7, 4)
#define AD7124_ID_DEVICE_ID_GET(x)   FIELD_GET(AD7124_ID_DEVICE_ID_MSK, x)
#define AD7124_ID_SILICON_REV_MSK    GENMASK(3, 0)
#define AD7124_ID_SILICON_REV_GET(x) FIELD_GET(AD7124_ID_SILICON_REV_MSK, x)

// Error register (addr=0x06, len=24, R)
#define AD7124_ERROR_LDO_CAP_ERR_MSK      BIT(19)
#define AD7124_ERROR_LDO_CAP_ERR(x)       FIELD_GET(AD7124_ERROR_LDO_CAP_ERR_MSK, x)
#define AD7124_ERROR_ADC_CAL_ERR_MSK      BIT(18)
#define AD7124_ERROR_ADC_CAL_ERR(x)       FIELD_GET(AD7124_ERROR_ADC_CAL_ERR_MSK, x)
#define AD7124_ERROR_ADC_CONV_ERR_MSK     BIT(17)
#define AD7124_ERROR_ADC_CONV_ERR(x)      FIELD_GET(AD7124_ERROR_ADC_CONV_ERR_MSK, x)
#define AD7124_ERROR_ADC_SAT_ERR_MSK      BIT(16)
#define AD7124_ERROR_ADC_SAT_ERR(x)       FIELD_GET(AD7124_ERROR_ADC_SAT_ERR_MSK, x)
#define AD7124_ERROR_AINP_OV_ERR_MSK      BIT(15)
#define AD7124_ERROR_AINP_OV_ERR(x)       FIELD_GET(AD7124_ERROR_AINP_OV_ERR_MSK, x)
#define AD7124_ERROR_AINP_UV_ERR_MSK      BIT(14)
#define AD7124_ERROR_AINP_UV_ERR(x)       FIELD_GET(AD7124_ERROR_AINP_UV_ERR_MSK, x)
#define AD7124_ERROR_AINM_OV_ERR_MSK      BIT(13)
#define AD7124_ERROR_AINM_OV_ERR(x)       FIELD_GET(AD7124_ERROR_AINM_OV_ERR_MSK, x)
#define AD7124_ERROR_AINM_UV_ERR_MSK      BIT(12)
#define AD7124_ERROR_AINM_UV_ERR(x)       FIELD_GET(AD7124_ERROR_AINM_UV_ERR_MSK, x)
#define AD7124_ERROR_REF_DET_ERR_MSK      BIT(11)
#define AD7124_ERROR_REF_DET_ERR(x)       FIELD_GET(AD7124_ERROR_REF_DET_ERR_MSK, x)
#define AD7124_ERROR_DLDO_PSM_ERR_MSK     BIT(9)
#define AD7124_ERROR_DLDO_PSM_ERR(x)      FIELD_GET(AD7124_ERROR_DLDO_PSM_ERR_MSK, x)
#define AD7124_ERROR_ALDO_PSM_ERR_MSK     BIT(7)
#define AD7124_ERROR_ALDO_PSM_ERR(x)      FIELD_GET(AD7124_ERROR_ALDO_PSM_ERR_MSK, x)
#define AD7124_ERROR_SPI_IGNORE_ERR_MSK   BIT(6)
#define AD7124_ERROR_SPI_IGNORE_ERR(x)    FIELD_GET(AD7124_ERROR_SPI_IGNORE_ERR_MSK, x)
#define AD7124_ERROR_SPI_SCLK_CNT_ERR_MSK BIT(5)
#define AD7124_ERROR_SPI_SCLK_CNT_ERR(x)  FIELD_GET(AD7124_ERROR_SPI_SCLK_CNT_ERR_MSK, x)
#define AD7124_ERROR_SPI_READ_ERR_MSK     BIT(4)
#define AD7124_ERROR_SPI_READ_ERR(x)      FIELD_GET(AD7124_ERROR_SPI_READ_ERR_MSK, x)
#define AD7124_ERROR_SPI_WRITE_ERR_MSK    BIT(3)
#define AD7124_ERROR_SPI_WRITE_ERR(x)     FIELD_GET(AD7124_ERROR_SPI_WRITE_ERR_MSK, x)
#define AD7124_ERROR_SPI_CRC_ERR_MSK      BIT(2)
#define AD7124_ERROR_SPI_CRC_ERR (x)      FIELD_GET(AD7124_ERROR_SPI_CRC_ERR_MSK, x)
#define AD7124_ERROR_MM_CRC_ERR_MSK       BIT(1)
#define AD7124_ERROR_MM_CRC_ERR (x)       FIELD_GET(AD7124_ERROR_MM_CRC_ERR_MSK, x)
#define AD7124_ERROR_ROM_CRC_ERR_MSK      BIT(0)
#define AD7124_ERROR_ROM_CRC_ERR (x)      FIELD_GET(AD7124_ERROR_ROM_CRC_ERR_MSK, x)

// Error enable register (addr=0x07, len=24, RW)
#define AD7124_ERROR_EN_MCLK_CNT_EN_MSK           BIT(22)
#define AD7124_ERROR_EN_MCLK_CNT_EN(x)            FIELD_GET(AD7124_ERROR_EN_MCLK_CNT_EN_MSK, x)
#define AD7124_ERROR_EN_LDO_CAP_CHK_TEST_EN_MSK   BIT(21)
#define AD7124_ERROR_EN_LDO_CAP_CHK_TEST_EN(x)    FIELD_GET(AD7124_ERROR_EN_LDO_CAP_CHK_TEST_EN_MSK, x)
#define AD7124_ERROR_EN_LDO_CAP_CHK_MSK           GENMASK(20, 19)
#define AD7124_ERROR_EN_LDO_CAP_CHK(x)            FIELD_GET(AD7124_ERROR_EN_LDO_CAP_CHK_MSK, x)
#define AD7124_ERROR_EN_ADC_CAL_ERR_EN_MSK        BIT(18)
#define AD7124_ERROR_EN_ADC_CAL_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_ADC_CAL_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_ADC_CONV_ERR_EN_MSK       BIT(17)
#define AD7124_ERROR_EN_ADC_CONV_ERR_EN(x)        FIELD_GET(AD7124_ERROR_EN_ADC_CONV_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_ADC_SAT_ERR_EN_MSK        BIT(16)
#define AD7124_ERROR_EN_ADC_SAT_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_ADC_SAT_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_AINP_OV_ERR_EN_MSK        BIT(15)
#define AD7124_ERROR_EN_AINP_OV_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_AINP_OV_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_AINP_UV_ERR_EN_MSK        BIT(14)
#define AD7124_ERROR_EN_AINP_UV_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_AINP_UV_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_AINM_OV_ERR_EN_MSK        BIT(13)
#define AD7124_ERROR_EN_AINM_OV_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_AINM_OV_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_AINM_UV_ERR_EN_MSK        BIT(12)
#define AD7124_ERROR_EN_AINM_UV_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_AINM_UV_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_REF_DET_ERR_EN_MSK        BIT(11)
#define AD7124_ERROR_EN_REF_DET_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_REF_DET_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_DLDO_PSM_TRIP_TEST_EN_MSK BIT(10)
#define AD7124_ERROR_EN_DLDO_PSM_TRIP_TEST_EN(x)  FIELD_GET(AD7124_ERROR_EN_DLDO_PSM_TRIP_TEST_EN_MSK, x)
#define AD7124_ERROR_EN_DLDO_PSM_ERR_EN_MSK       BIT(9)
#define AD7124_ERROR_EN_DLDO_PSM_ERR_EN(x)        FIELD_GET(AD7124_ERROR_EN_DLDO_PSM_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_ALDO_PSM_TRIP_TEST_EN_MSK BIT(8)
#define AD7124_ERROR_EN_ALDO_PSM_TRIP_TEST_EN(x)  FIELD_GET(AD7124_ERROR_EN_ALDO_PSM_TRIP_TEST_EN_MSK, x)
#define AD7124_ERROR_EN_ALDO_PSM_ERR_EN_MSK       BIT(7)
#define AD7124_ERROR_EN_ALDO_PSM_ERR_EN(x)        FIELD_GET(AD7124_ERROR_EN_ALDO_PSM_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_SPI_IGNORE_ERR_EN_MSK     BIT(6)
#define AD7124_ERROR_EN_SPI_IGNORE_ERR_EN(x)      FIELD_GET(AD7124_ERROR_EN_SPI_IGNORE_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_SPI_SCLK_CNT_ERR_EN_MSK   BIT(5)
#define AD7124_ERROR_EN_SPI_SCLK_CNT_ERR_EN(x)    FIELD_GET(AD7124_ERROR_EN_SPI_SCLK_CNT_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_SPI_READ_ERR_EN_MSK       BIT(4)
#define AD7124_ERROR_EN_SPI_READ_ERR_EN(x)        FIELD_GET(AD7124_ERROR_EN_SPI_READ_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_SPI_WRITE_ERR_EN_MSK      BIT(3)
#define AD7124_ERROR_EN_SPI_WRITE_ERR_EN(x)       FIELD_GET(AD7124_ERROR_EN_SPI_WRITE_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_SPI_CRC_ERR_EN_MSK        BIT(2)
#define AD7124_ERROR_EN_SPI_CRC_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_SPI_CRC_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_MM_CRC_ERR_EN_MSK         BIT(1)
#define AD7124_ERROR_EN_MM_CRC_ERR_EN(x)          FIELD_GET(AD7124_ERROR_EN_MM_CRC_ERR_EN_MSK, x)
#define AD7124_ERROR_EN_ROM_CRC_ERR_EN_MSK        BIT(0)
#define AD7124_ERROR_EN_ROM_CRC_ERR_EN(x)         FIELD_GET(AD7124_ERROR_EN_ROM_CRC_ERR_EN_MSK, x)

// MCLK count register (addr=0x08, len=8, R)
#define AD7124_MCLK_COUNT_MSK GENMASK(7, 0)
#define AD7124_MCLK_COUNT     FIELD_GET(AD7124_MCLK_COUNT_MSK, x)

// CHANNEL registers 0 to 15 (addr=0x09~0x18, len=16, RW)
#define AD7124_CHANNEL_EN_MSK      BIT(15)
#define AD7124_CHANNEL_EN(x)       FIELD_PREP(AD7124_CHANNEL_EN_MSK, x)
#define AD7124_CHANNEL_SETUP_MSK   GENMASK(14, 12)
#define AD7124_CHANNEL_SETUP(x)    FIELD_PREP(AD7124_CHANNEL_SETUP_MSK, x)
#define AD7124_CHANNEL_AINP_MSK    GENMASK(9, 5)
#define AD7124_CHANNEL_AINP(x)     FIELD_PREP(AD7124_CHANNEL_AINP_MSK, x)
#define AD7124_CHANNEL_AINM_MSK    GENMASK(4, 0)
#define AD7124_CHANNEL_AINM(x)     FIELD_PREP(AD7124_CHANNEL_AINM_MSK, x)

// CONFIG registers 0 to 7 (addr=0x19~0x20, len=16, RW)
#define AD7124_CONFIG_BIPOLAR_MSK  BIT(11)
#define AD7124_CONFIG_BIPOLAR(x)   FIELD_PREP(AD7124_CONFIG_BIPOLAR_MSK, x)
#define AD7124_CONFIG_BURNOUT_MSK  GENMASK(10, 9)
#define AD7124_CONFIG_BURNOUT(x)   FIELD_PREP(AD7124_CONFIG_BURNOUT_MSK, x)
#define AD7124_CONFIG_REF_BUFP_MSK BIT(8)
#define AD7124_CONFIG_REF_BUFP(x)  FIELD_PREP(AD7124_CONFIG_REF_BUFP_MSK, x)
#define AD7124_CONFIG_REF_BUFM_MSK BIT(7)
#define AD7124_CONFIG_REF_BUFM(x)  FIELD_PREP(AD7124_CONFIG_REF_BUFM_MSK, x)
#define AD7124_CONFIG_AIN_BUFP_MSK BIT(6)
#define AD7124_CONFIG_AIN_BUFP(x)  FIELD_PREP(AD7124_CONFIG_AIN_BUFP_MSK, x)
#define AD7124_CONFIG_AIN_BUFM_MSK BIT(5)
#define AD7124_CONFIG_AIN_BUFM(x)  FIELD_PREP(AD7124_CONFIG_AIN_BUFM_MSK, x)
#define AD7124_CONFIG_REF_SEL_MSK  GENMASK(4, 3)
#define AD7124_CONFIG_REF_SEL(x)   FIELD_PREP(AD7124_CONFIG_REF_SEL_MSK, x)
#define AD7124_CONFIG_PGA_MSK      GENMASK(2, 0)
#define AD7124_CONFIG_PGA(x)       FIELD_PREP(AD7124_CONFIG_PGA_MSK, x)
#define AD7124_CONFIG_IN_BUFF_MSK  GENMASK(6, 5)
#define AD7124_CONFIG_IN_BUFF(x)   FIELD_PREP(AD7124_CONFIG_IN_BUFF_MSK, x)

// FILTER registers 0 to 7 (addr=0x21~0x28, len=24, RW)
#define AD7124_FILTER_FILTER_MSK       GENMASK(23, 21)
#define AD7124_FILTER_FILTER(x)        FIELD_PREP(AD7124_FILTER_FILTER_MSK, x)
#define AD7124_FILTER_REJ60_MSK        BIT(20)
#define AD7124_FILTER_REJ60(x)         FIELD_PREP(AD7124_FILTER_REJ60_MSK, x)
#define AD7124_FILTER_POST_FILTER_MSK  GENMASK(19, 17)
#define AD7124_FILTER_POST_FILTER(x)   FIELD_PREP(AD7124_FILTER_POST_FILTER_MSK, x)
#define AD7124_FILTER_SINGLE_CYCLE_MSK BIT(16)
#define AD7124_FILTER_SINGLE_CYCLE(x)  FIELD_PREP(AD7124_FILTER_SINGLE_CYCLE_MSK, x)
#define AD7124_FILTER_FS_MSK           GENMASK(10, 0)
#define AD7124_FILTER_FS(x)            FIELD_PREP(AD7124_FILTER_FS_MSK, x)


LOG_MODULE_REGISTER(adc);

typedef enum {
    AD7124_CONFIG_IDX_0 = 0,
    AD7124_CONFIG_IDX_1,
    AD7124_CONFIG_IDX_2,
    AD7124_CONFIG_IDX_3,
    AD7124_CONFIG_IDX_4,
    AD7124_CONFIG_IDX_5,
    AD7124_CONFIG_IDX_6,
    AD7124_CONFIG_IDX_7
} ad7124_config_idx_t;

typedef enum {
    AD7124_CONFIG_PGA_1 = 0,
    AD7124_CONFIG_PGA_2,
    AD7124_CONFIG_PGA_4,
    AD7124_CONFIG_PGA_8,
    AD7124_CONFIG_PGA_16,
    AD7124_CONFIG_PGA_32,
    AD7124_CONFIG_PGA_64,
    AD7124_CONFIG_PGA_128
} ad7124_config_pga_t;

typedef enum {
    AD7124_OP_MODE_CONT = 0, // continuous conversion mode (default)
    AD7124_OP_MODE_SINGLE,
    AD7124_OP_MODE_STANDBY,
    AD7124_OP_MODE_POWER_DOWN,
    AD7124_OP_MODE_IDLE,
    AD7124_OP_MODE_INTERNAL_ZERO_SCALE,
    AD7124_OP_MODE_INTERNAL_FULL_SCALE,
    AD7124_OP_MODE_SYSTEM_ZERO_SCALE,
    AD7124_OP_MODE_SYSTEM_FULL_SCALE,
    AD7124_OP_MODE_RESERVED
} ad7124_op_mode_t;

typedef enum {
    AD7124_FILTER_IDX_0 = 0,
    AD7124_FILTER_IDX_1,
    AD7124_FILTER_IDX_2,
    AD7124_FILTER_IDX_3,
    AD7124_FILTER_IDX_4,
    AD7124_FILTER_IDX_5,
    AD7124_FILTER_IDX_6,
    AD7124_FILTER_IDX_7
} ad7124_filter_idx_t;

typedef enum {
    AD7124_GAIN_IDX_0 = 0,
    AD7124_GAIN_IDX_1,
    AD7124_GAIN_IDX_2,
    AD7124_GAIN_IDX_3,
    AD7124_GAIN_IDX_4,
    AD7124_GAIN_IDX_5,
    AD7124_GAIN_IDX_6,
    AD7124_GAIN_IDX_7
} ad7124_gain_idx_t;

typedef enum {
    AD7124_OFFSET_IDX_0 = 0,
    AD7124_OFFSET_IDX_1,
    AD7124_OFFSET_IDX_2,
    AD7124_OFFSET_IDX_3,
    AD7124_OFFSET_IDX_4,
    AD7124_OFFSET_IDX_5,
    AD7124_OFFSET_IDX_6,
    AD7124_OFFSET_IDX_7
} ad7124_offset_idx_t;

typedef enum {
    AD7124_POWER_MODE_FULL = 0,
    AD7124_POWER_MODE_MID,
    AD7124_POWER_MODE_LOW
} ad7124_power_mode_t;

typedef struct {
    bool    bipolar;
    uint8_t burnout;
    bool    refBufp;
    bool    refBufm;
    bool    ainBufp;
    bool    ainBufm;
    uint8_t refSel;
    uint8_t pga;
} ad7124_config_t;

typedef struct {
    bool    rdy;
    bool    errorFlag;
    bool    porFlag;
    uint8_t chActive;
} ad7124_status_t;

// SPI 
#define ADCSPI_NODE DT_NODELABEL(adcspi)
static const struct device* _gSpiDev = DEVICE_DT_GET(ADCSPI_NODE);
static const struct spi_cs_control _gSpiCs = {
    .gpio = SPI_CS_GPIOS_DT_SPEC_GET(DT_NODELABEL(reg_adcspi)),
    .delay = 0,
};
static const struct spi_config _gSpiCfg = {
    .operation = SPI_WORD_SET(8)  |
                 SPI_TRANSFER_MSB |
                 SPI_MODE_CPOL    |
                 SPI_MODE_CPHA,
    .frequency = 8000000,
    .slave = 0,
    .cs = &_gSpiCs,
};

static int _ad7124GetFclk(const ad7124_power_mode_t mode) {
    int fclk = 0;
    switch (mode) {
        case AD7124_POWER_MODE_FULL:
            fclk = AD7124_FCLK_FULL_POWER;
            break;

        case AD7124_POWER_MODE_MID:
            fclk = AD7124_FCLK_MID_POWER;
            break;

        case AD7124_POWER_MODE_LOW:
            fclk = AD7124_FCLK_LOW_POWER;
            break;
    }

    return fclk;
}

static float _ad7124GetF3dB(const int fAdc) {
    const int f3dB = 0.23 * fAdc;

    return f3dB;
}

static uint16_t _ad7124GetRegFilterFS(const ad7124_power_mode_t mode, const int fAdc) {
    const int fClk    = _ad7124GetFclk(mode);
    const uint16_t fs = fClk/(32 * fAdc) + 0.5;

    return fs;
}

static uint8_t _ad7124GetRegConfigAddr(const ad7124_config_idx_t idx) {
    const uint8_t regAddr = AD7124_REG_ADDR_CONFIG(idx);

    return regAddr;
}

static uint8_t _ad7124GetRegFilterAddr(const ad7124_filter_idx_t idx) {
    const uint8_t regAddr = AD7124_REG_ADDR_FILTER(idx);

    return regAddr;
}

static uint8_t _ad7124GetRegGainAddr(const ad7124_gain_idx_t idx) {
    const uint8_t regAddr = AD7124_REG_ADDR_GAIN(idx);

    return regAddr;
}

static uint8_t _ad7124GetRegOffsetAddr(const ad7124_offset_idx_t idx) {
    const uint8_t regAddr = AD7124_REG_ADDR_OFFSET(idx);

    return regAddr;
}

static ad7124_status_t _ad7124Uint8ToStatus(const uint8_t u8) {
    ad7124_status_t status;
    status.rdy       = AD7124_STATUS_RDY(u8);
    status.errorFlag = AD7124_STATUS_ERROR_FLAG(u8);
    status.porFlag   = AD7124_STATUS_POR_FLAG(u8);
    status.chActive  = AD7124_STATUS_CH_ACTIVE(u8);
}


static int _ad7124RegRead(const uint8_t regAddr, const size_t len, uint32_t* regVal) {
    int ret;
    uint8_t txBuf[1] = { regAddr };

    const struct spi_buf txSpiBuf = {
        .buf = txBuf,
        .len = sizeof(txBuf)
    };

    const struct spi_buf_set txSpiBufSet = {
        .buffers = &txSpiBuf,
        .count   = 1
    };

    uint8_t rxBufU8[4] = {0};
    struct spi_buf rxSpiBuf = {
        .buf = rxBufU8,
        .len = sizeof(rxBufU8)
    };
    const struct spi_buf_set rxSpiBufSet = { .buffers = &rxSpiBuf, .count = 1 };

    ret = spi_transceive(_gSpiDev, &_gSpiCfg, &txSpiBufSet, &rxSpiBufSet);
    if (ret < 0) {
        LOG_ERR("Reading AD7124 register failed, regAddr=0x%x, len=%u, err=%d\n", regAddr, len, ret);
        return ERROR_ADC_READ_FAIL;
    }

    LOG_DBG("%x %x %x\n", rxBufU8[3], rxBufU8[2], rxBufU8[1]);
    *regVal = spiData2Uint32(rxBufU8);

    return 0;
}

static int _ad7124RegStausRead(ad7124_status_t* status) {
    uint32_t regVal;
    if (_ad7124RegRead(AD7124_REG_ADDR_STATUS, AD7124_REG_LEN_STATUS, &regVal) < 0) {
        LOG_ERR("Reading AD7124 status register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    *status = _ad7124Uint8ToStatus((uint8_t) regVal);

    return 0;
}

static int _ad7124RegAdcCtrlRead(uint16_t* adcCtrl) {
    uint32_t regVal;
    if (_ad7124RegRead(AD7124_REG_ADDR_ADC_CONTROL, AD7124_REG_LEN_ADC_CTRL, &regVal) < 0) {
        LOG_ERR("Reading AD7124 ADC_CONTROL register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    *adcCtrl = (uint16_t) regVal;

    return 0;
}

static int _ad7124RegIdRead(uint8_t* id) {
    if (_ad7124RegRead(AD7124_REG_ADDR_ID, AD7124_REG_LEN_ID, id) < 0) {
        LOG_ERR("Reading AD7124 ID register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    return 0;
}

static int _ad7124RegIdCheck(void) {
    uint8_t id;
    if (_ad7124RegIdRead(&id) < 0) {
        LOG_ERR("Reading AD7124 ID register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    if (AD7124_ID_DEVICE_ID_GET(id) != AD7124_4_CHIPID) {
        LOG_ERR("AD7124 ID register incorrect id=0x%x\n", id);
        return ERROR_ADC_READ_FAIL;
    }

    return 0;
}

static ad7124_config_t _ad7124U32ToConfig(const uint32_t val) {
    ad7124_config_t cfg;
    cfg.bipolar = (val >> AD7124_REG_CONFIG_BIT_BIPOLAR)  & 0x00000001;
    cfg.burnout = (val >> AD7124_REG_CONFIG_BIT_BURNOUT)  & 0x00000003;
    cfg.refBufp = (val >> AD7124_REG_CONFIG_BIT_REF_BUFP) & 0x00000001;
    cfg.refBufm = (val >> AD7124_REG_CONFIG_BIT_REF_BUFM) & 0x00000001;
    cfg.ainBufp = (val >> AD7124_REG_CONFIG_BIT_AIN_BUFP) & 0x00000001;
    cfg.ainBufm = (val >> AD7124_REG_CONFIG_BIT_AIN_BUFM) & 0x00000001;
    cfg.refSel  = (val >> AD7124_REG_CONFIG_BIT_REF_SEL)  & 0x00000003;
    cfg.pga     = val & 0x00000007;

    return cfg;

}

static uint32_t _ad7124ConfigToU32(const ad7124_config_t cfg) {
    uint32_t val;
    val = AD7124_CONFIG_BIPOLAR(cfg.bipolar)  |
          AD7124_CONFIG_BURNOUT(cfg.burnout)  |
          AD7124_CONFIG_REF_BUFP(cfg.refBufp) |
          AD7124_CONFIG_REF_BUFM(cfg.refBufm) |
          AD7124_CONFIG_AIN_BUFP(cfg.ainBufp) |
          AD7124_CONFIG_AIN_BUFM(cfg.ainBufm) |
          AD7124_CONFIG_REF_SEL(cfg.refSel)   |
          AD7124_CONFIG_PGA(cfg.pga);

    return val;
}

static int _ad7124RegConfigRead(const ad7124_config_idx_t idx, ad7124_config_t* config) {
    uint32_t regVal;
    const uint8_t regAddr = _ad7124GetRegConfigAddr(idx);
    if (_ad7124RegRead(regAddr, AD7124_REG_LEN_CONFIG, &regVal) < 0) {
        LOG_ERR("Reading AD7124 config register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    return 0;
}

static int _ad7124RegFilterRead(const ad7124_filter_idx_t idx, uint32_t* regVal) {
    const uint8_t regAddr = _ad7124GetRegFilterAddr(idx);
    if (_ad7124RegRead(regAddr, AD7124_REG_LEN_FILTER, regVal) < 0) {
        LOG_ERR("Reading AD7124 filter register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    return 0;
}

static int _ad7124RegOffsetRead(const ad7124_offset_idx_t idx, uint32_t* regVal) {
    const uint8_t regAddr = _ad7124GetRegOffsetAddr(idx);
    if (_ad7124RegRead(regAddr, AD7124_REG_LEN_OFFSET, regVal) < 0) {
        LOG_ERR("Reading AD7124 offset register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    return 0;
}

static int _ad7124RegGainRead(const ad7124_gain_idx_t idx, uint32_t* regVal) {
    const uint8_t regAddr = _ad7124GetRegGainAddr(idx);
    if (_ad7124RegRead(regAddr, AD7124_REG_LEN_GAIN, regVal) < 0) {
        LOG_ERR("Reading AD7124 gain register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    return 0;
}

static int _ad7124RegWrite(uint8_t regAddr, const uint32_t regVal, const size_t len) {
    int ret;
    uint8_t data[len];
    for (int i=0; i<len; i++) {
        data[i] = (regVal >> i) & 0xFF;
    }

    uint8_t wbuf[len+1];
    wbuf[0] = regAddr;
    for (int i=0; i<len; i++) {
        wbuf[i+1] = data[i];
    }
    const struct spi_buf     txSpiBuf = { .buf = wbuf, .len = sizeof(wbuf)};
    const struct spi_buf_set txSpiBufSet = {.buffers = &txSpiBuf, .count = 1};

    ret = spi_write(_gSpiDev, &_gSpiCfg, &txSpiBufSet);
    if (ret < 0) {
        LOG_ERR("Writing LIS2DH12 register failed, regAddr=0x%x, regVal=0x%x, err=%d\n", regAddr, regVal, ret);
        return ERROR_GSENSOR_WRITE_FAIL;
    }

    return 0;
}

static int _ad7124RegWriteMask(const uint8_t regAddr, unsigned int mask, unsigned int val, unsigned int len) {
    unsigned int regVal;
    int ret;

    ret = _ad7124RegRead(regAddr, len, &regVal);
    if (ret < 0) {
        return ret;
    }

    regVal &= ~mask;
    regVal |= val;

    return _ad7124RegWrite(regAddr, regVal, len);
}

static int _ad7124RegCommsWrite(const uint32_t regVal) {
    if (_ad7124RegWrite(AD7124_REG_ADDR_COMMS, regVal, AD7124_REG_LEN_COMMS) < 0) {
        LOG_ERR("Writing AD7124 comms register failed\n");
        return ERROR_ADC_WRITE_FAIL;
    }

    return 0;
}

static int _ad7124RegAdcCtrlWriteMode(const ad7124_op_mode_t mode) {
    uint16_t regVal;

    if (_ad7124RegAdcCtrlRead(&regVal) < 0) {
        LOG_ERR("Reading AD7124 ADC_CONTROL register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    regVal &= 0xFFC3;
    regVal |= mode;

    if (_ad7124RegWrite(AD7124_REG_ADDR_COMMS, regVal, AD7124_REG_LEN_COMMS) < 0) {
        LOG_ERR("Writing AD7124 comms register failed\n");
        return ERROR_ADC_WRITE_FAIL;
    }

    return 0;
}

static int _ad7124RegConfigWrite(const ad7124_config_idx_t idx, const ad7124_config_t cfg) {
    const int regAddr     = _ad7124GetRegConfigAddr(idx);
    const uint32_t regVal = _ad7124ConfigToU32(cfg);

    if (_ad7124RegWrite(regAddr, regVal, AD7124_REG_LEN_FILTER) < 0)  {
        LOG_ERR("Writing AD7124 config register failed\n");
        return ERROR_ADC_WRITE_FAIL;
    }

    return 0;
}

static int _ad7124RegConfigWritePga(const ad7124_config_idx_t idx, const ad7124_config_pga_t pga) {
    const int regAddr = _ad7124GetRegConfigAddr(idx);
    ad7124_config_t cfg;
    if (_ad7124RegConfigRead(idx, &cfg) < 0) {
        LOG_ERR("Writing AD7124 config register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    uint32_t regVal = _ad7124ConfigToU32(cfg);
    regVal &= 0xFFF800;
    regVal |= pga;

    if (_ad7124RegWrite(regAddr, regVal, AD7124_REG_LEN_FILTER) < 0)  {
        LOG_ERR("Writing AD7124 config register failed\n");
        return ERROR_ADC_WRITE_FAIL;
    }

    return 0;
}

static int _ad7124RegFilterWriteFs(const ad7124_filter_idx_t idx, const uint16_t fs) {
    const int regAddr = _ad7124GetRegFilterAddr(idx);
    uint32_t regVal;
    if (_ad7124RegFilterRead(idx, &regVal) < 0) {
        LOG_ERR("Reading AD7124 filter register failed\n");
        return ERROR_ADC_READ_FAIL;
    }

    regVal &= 0xFFFFF8;
    regVal |= fs;

    if (_ad7124RegWrite(regAddr, regVal, AD7124_REG_LEN_FILTER) < 0) {
        LOG_ERR("Writing AD7124 filter register failed\n");
        return ERROR_ADC_WRITE_FAIL;
    }

    return 0;
}

static int _ad7124RegOffsetWrite(const ad7124_offset_idx_t idx, const uint32_t offset) {
    const int regAddr = _ad7124GetRegOffsetAddr(idx);

    if (_ad7124RegWrite(regAddr, offset, AD7124_REG_LEN_OFFSET) < 0) {
        LOG_ERR("Writing AD7124 offset register failed\n");
        return ERROR_ADC_WRITE_FAIL;
    }

    return 0;
}

static int _ad7124RegGainWrite(const ad7124_gain_idx_t idx, const uint32_t gain) {
    const int regAddr = _ad7124GetRegOffsetAddr(idx);

    if (_ad7124RegWrite(regAddr, gain, AD7124_REG_LEN_GAIN) < 0) {
        LOG_ERR("Writing AD7124 offset register failed\n");
        return ERROR_ADC_WRITE_FAIL;
    }

    return 0;
}

int ad7124Init(void) {
    if (!device_is_ready(_gSpiDev)) {
        printk("lis2dh12 device isn't ready!\n");
        return ERROR_GSENSOR_INITIALIZE_FAIL;
    }

    if(!device_is_ready(_gSpiCs.gpio.port)){
        printk("lis2dh12 chip select device isn't ready!\n");
        return ERROR_GSENSOR_INITIALIZE_FAIL;
    }

    if (_ad7124RegIdCheck() < 0) {
        return ERROR_ADC_READ_FAIL;
    }

    return 0;
}

#endif